#include<cstdio>
#include<cstdlib>
#include<cstring>
#include<cmath>

///////////////////////////////////////
//
// WRITTEN BY MARK GOLDMAN
// U.S. GEOLOGICAL SURVEY
// MS 977
// 345 MIDDLEFIELD RD
// MENLO PARK, CALIFORNIA 94025
// goldman@usgs.gov
//
///////////////////////////////////////

void rsy2Segy(int argc, char** argv);

void getValues(const char* input, int& sampIntSave, int& numSampSave, 
   int& integerFormat, float& scaling);

void writeReelHeader(FILE* out, int sampIntSave, int numSampSave);

void writeSegy(const char* input, FILE* out, 
   int sampIntSave, int numSampSave, int ffid, int integerFormat, float scaling);

float ibm2ieee(const float ibm);
double ibm2ieee(const double ibm);

main(int argc, char** argv) {
   rsy2Segy(argc, argv);
}

void rsy2Segy(int argc, char** argv) {

   if (argc < 4) {
      printf("\n");
      printf("%s reads one or more Passcal .RSY Segy files and\n",argv[0]);
      printf("converts them to a standard Segy file by adding a 3600 byte reel header\n");
      printf("(samples are written in IEEE floating point format)\n");
      printf("\n");
      printf("may use wildcards to indicate input filenames\n");
      printf("\n");
      printf("each new input file will increment the ffid by 1\n");
      printf("\n");
      printf("%s assumes NO 3600 byte reel header for input .RSY files\n",argv[0]);
      printf("\n");
      printf("sample values must be in either 2 or 4 byte integer format\n");
      printf("(as determined by bytes 205-206)\n");
      printf("\n");
      printf("sample values multiplied by scaling factor, defined in bytes 221-224\n");
      printf("\n");
      printf("number of samples read from bytes 229-232 (32767 samples/trace max)\n");
      printf("\n");
      printf("WARNING:  all input files must have the same\n");
      printf("          sample rate and number of samples\n");
      printf("\n");
      printf("format:  %s  output_segy_filename  starting_ffid  input_rsy_filename[s]\n", argv[0]);
      printf("\n");
      printf("example: %s  output  1001  input*\n",argv[0]);
      printf("\n");
      exit(1);
   }

/////////////////////////////////////////

   // RUNTIME PARAMETERS
   int sampIntSave;
   int numSampSave;
   int integerFormat;
   float scaling;

   // GET ARGUMENTS
   char output[200];
   strcpy(output, argv[1]);

   int ffid = atoi(argv[2]);
   if (ffid < 1) {
      printf("\nERROR:  starting_ffid < 1\n\n");
      exit(1);
   }

   // OPEN OUTPUT FILE
   FILE* out;
   if ((out = fopen(output,"w")) == NULL) {
      printf("\nERROR:  cannot find file %s\n\n",output);
      exit(1);
   }

   // GET SAMPLE INTERVAL AND NUMBER OF SAMPLES
   getValues(argv[3], sampIntSave, numSampSave, integerFormat, scaling);
   printf("\n");
   printf("%6d = sample interval in microseconds\n", sampIntSave);
   printf("%6d = number of samples per trace\n", numSampSave);
   printf("%6d = integer format of samples (0 = 2 bytes, 1 = 4 bytes)\n", integerFormat);
   printf("%12.4f = scaling to multiply samples by\n", scaling);
   printf("\n");

   // WRITE 3600 BYTE REEL HEADER
   writeReelHeader(out, sampIntSave, numSampSave);

   // WRITE SEGY FILES
   for (int i = 3; i < argc; i++) {
      writeSegy(argv[i], out, sampIntSave, numSampSave, ffid, integerFormat, scaling);
      ffid++;
   }

   // CLOSE FILES
   fclose(out);

}

/////////////////////////////////////////

void getValues(const char* input, int& sampIntSave, int& numSampSave, 
   int& integerFormat, float& scaling) {

   // OPEN INPUT FILE
   FILE* in;
   if ((in = fopen(input,"r")) == NULL) {
      printf("\nERROR:  cannot open file %s\n\n",input);
      exit(1);
   }

   // READ TRACE HEADER
   char traceHeader[240];
   if (fread(&traceHeader, 240, 1, in) != 1) {
      printf("\nERROR:  cannot read file %s\n\n",input);
      exit(1);
   }

   // GET NUM_SAMP, SAMP_INT FROM TRACE HEADER VARIABLE
   union {
      char c[2];
      short s;
   } test2;

   union {
      char c[4];
      int i;
      float f;
   } test4;

   test2.c[0] = traceHeader[116];
   test2.c[1] = traceHeader[117];
   sampIntSave = (int)test2.s;
   if (sampIntSave > 32767) {
      printf("\nERROR:  more than 32767 samples per trace\n\n");
      exit(1);
   }

   test2.c[0] = traceHeader[205];
   test2.c[1] = traceHeader[206];
   integerFormat = (int)test2.s;

   test4.c[0] = traceHeader[221];
   test4.c[1] = traceHeader[222];
   test4.c[2] = traceHeader[223];
   test4.c[3] = traceHeader[224];
   scaling = test4.f;

   test4.c[0] = traceHeader[229];
   test4.c[1] = traceHeader[230];
   test4.c[2] = traceHeader[231];
   test4.c[3] = traceHeader[232];
   numSampSave = test4.i;

   // CLOSE INPUT FILE
   fclose(in);

}

/////////////////////////////////////////

void writeReelHeader(FILE* out, int sampIntSave, int numSampSave) {

   // REEL HEADER PARAMETERS
   int   rjobid = 1;
   int   rlineNumber = 1;
   int   rreelNumber = 0;
   short rnumTraces = 0;
   short rnumAuxTraces = 0;
   short rsampInt;
   short rsampIntOriginal;
   short rnumSamp;
   short rnumSampOriginal;
   short rformat = 5;
   short rfold = 0;
   short rsorting = 1;
   short rnumSummed = 0;
   short rsweepStart = 0;
   short rsweepEnd = 0;
   short rsweepLength = 0;
   short rsweepType = 0;
   short rsweepChannel = 0;
   short rsweepTaperStart = 0;
   short rsweepTaperEnd = 0;
   short rsweepTaperType = 0;
   short rcorrelated = 2;
   short rgain = 1;
   short ramplitude = 4;
   short rsystem = 1;
   short rimpulse = 1;
   short rvibratory = 0;
   char  rdummy[340];

   // CREATE DUMMY LINE FOR 3200 BYTE REEL HEADER
   char dummyReel[80];
   dummyReel[0] = 195;
   for (int i = 1; i < 80; i++)
      dummyReel[i] = 64;

   // WRITE 3200 BYTE REEL HEADER USING DUMMY LINE
   for (int i = 0; i < 40; i++)
      fwrite(&dummyReel, 80, 1, out);

   // GET VALUES FOR 400 BYTE REEL HEADER
   rsampInt = (short)sampIntSave;
   rsampIntOriginal = (short)sampIntSave;

   if (numSampSave > 32767) {
      rnumSamp = 32767;
      rnumSampOriginal = 32767;
   }
   else {
      rnumSamp = (short)numSampSave;
      rnumSampOriginal = (short)numSampSave;
   }

   // WRITE 400 BYTE REEL HEADER
   fwrite(&rjobid,           4, 1, out);
   fwrite(&rlineNumber,      4, 1, out);
   fwrite(&rreelNumber,      4, 1, out);
   fwrite(&rnumTraces,       2, 1, out);
   fwrite(&rnumAuxTraces,    2, 1, out);
   fwrite(&rsampInt,         2, 1, out);
   fwrite(&rsampIntOriginal, 2, 1, out);
   fwrite(&rnumSamp,         2, 1, out);
   fwrite(&rnumSampOriginal, 2, 1, out);
   fwrite(&rformat,          2, 1, out);
   fwrite(&rfold,            2, 1, out);
   fwrite(&rsorting,         2, 1, out);
   fwrite(&rnumSummed,       2, 1, out);
   fwrite(&rsweepStart,      2, 1, out);
   fwrite(&rsweepEnd,        2, 1, out);
   fwrite(&rsweepLength,     2, 1, out);
   fwrite(&rsweepType,       2, 1, out);
   fwrite(&rsweepChannel,    2, 1, out);
   fwrite(&rsweepTaperStart, 2, 1, out);
   fwrite(&rsweepTaperEnd,   2, 1, out);
   fwrite(&rsweepTaperType,  2, 1, out);
   fwrite(&rcorrelated,      2, 1, out);
   fwrite(&rgain,            2, 1, out);
   fwrite(&ramplitude,       2, 1, out);
   fwrite(&rsystem,          2, 1, out);
   fwrite(&rimpulse,         2, 1, out);
   fwrite(&rvibratory,       2, 1, out);
   fwrite(&rdummy,         340, 1, out);

}

//////////////////////////////////////////////////////////

void writeSegy(const char* input, FILE* out, 
   int sampIntSave, int numSampSave, int ffid, int integerFormat, float scaling) {

////////////////////////////////
// TEST WITH STANDARD SEGY FILE
// char dummy[3600];
// fread(&dummy, 3600, 1, in);
////////////////////////////////

   char traceHeader[240];

   int numSamp;
   int sampInt;

   // OPEN INPUT FILE
   FILE* in;
   if ((in = fopen(input,"r")) == NULL) {
      printf("\nERROR:  cannot open file %s\n\n",input);
      exit(1);
   }

   // READ TRACES
   int runflag = 1;
   while (runflag) {

      // READ TRACE HEADER
      if (fread(&traceHeader, 240, 1, in) != 1) {
         runflag = 0;
         continue;
      }

      // GET NUM_SAMP, SAMP_INT FROM TRACE HEADER VARIABLE
      union {
         char c[2];
         short s;
      } test2;

      test2.c[0] = traceHeader[114];
      test2.c[1] = traceHeader[115];
      numSamp = test2.s;

      test2.c[0] = traceHeader[116];
      test2.c[1] = traceHeader[117];
      sampInt = (int)test2.s;

      // TEST NUM_SAMP, SAMP_INT
      if (sampInt != sampIntSave) {
         printf("\nERROR:  change in sample interval\n\n");
         exit(1);
      }

      if (numSamp != numSampSave) {
         printf("\nERROR:  change in number of samples\n\n");
         exit(1);
      }

      if (numSamp < 1) {
         printf("\nERROR:  number of samples < 1\n");
         exit(1);
      }

      // WRITE TRACE HEADER, PUTTING IN FFID
      char* temp = traceHeader + 12;
      fwrite(&traceHeader,   8, 1, out);
      fwrite(&ffid,          4, 1, out);
      fwrite(temp,         228, 1, out);

      // WRITE DATA VALUES AS IEEE VALUES
      short svalue;
      int   ivalue;
      float fvalue;

      for (int i = 0 ; i < numSamp; i++) {

         if (integerFormat == 0) {
            if (fread(&svalue, 2, 1, in) != 1) {
               printf("\nERROR:  unexpected end of segy file %s\n\n",input);
               exit(1);
            }
            ivalue = (int)svalue;
         }
         else {
            if (fread(&ivalue, 4, 1, in) != 1) {
               printf("\nERROR:  unexpected end of segy file %s\n\n",input);
               exit(1);
            }
         }

         fvalue = (float)ivalue * scaling;

         fwrite(&fvalue, 4, 1, out);

      }

   // CONTINUE TO NEXT TRACE
   }

   fclose(in);

}

//////////////////////////////////////////////////////////

float ibm2ieee(const float ibm) {

   // DEFINE AND SETUP UNION VARIABLES
   union value {
      float f;
      unsigned char c[4];
   };
   value src;

   // ASSIGN PARAMETER TO UNION
   src.f = ibm;

   // CONVERT TO FLOAT
   long IntMantissa =
        ((long)src.c[1] << 16)
      + ((long)src.c[2] << 8)
      + src.c[3];

   float  Mantissa  = float(IntMantissa) / float(0x1000000);
   float  PosResult = Mantissa * pow( 16.0, double((src.c[0] & 0x7F) - 64) );

   if (src.c[0] & 0x80)
      return -PosResult;
   else
      return PosResult;

}

double ibm2ieee(const double ibm) {

   // DEFINE AND SETUP UNION VARIABLES
   union value {
      double f;
      unsigned char c[8];
   };
   value src;

   // ASSIGN PARAMETER TO UNION
   src.f = ibm;

   // CONVERT TO FLOAT
   long long IntMantissa =
        ((long long)src.c[1] << 48)
      + ((long long)src.c[2] << 40)
      + ((long long)src.c[3] << 32)
      + ((long long)src.c[4] << 24)
      + ((long long)src.c[5] << 16)
      + ((long long)src.c[6] << 8)
      + src.c[7];

   double  Mantissa  = double(IntMantissa) / double(0x1000000);
   double  PosResult = Mantissa * pow( 16.0, double((src.c[0] & 0x7F) - 64) );

   if (src.c[0] & 0x80)
      return -PosResult;
   else
      return PosResult;

}

