#include<cstdio>
#include<cstdlib>
#include<cstring>
#include<cmath>

///////////////////////////////////////
//
// WRITTEN BY MARK GOLDMAN
// U.S. GEOLOGICAL SURVEY
// MS 977
// 345 MIDDLEFIELD RD
// MENLO PARK, CALIFORNIA 94025
// goldman@usgs.gov
//
///////////////////////////////////////

main(int argc, char** argv) {
 
   void mcread(char*  dummy, const int nbytes, const int nitems, FILE* file);
   void msread(short* dummy, const int nbytes, const int nitems, FILE* file);
   void miread(int*   dummy, const int nbytes, const int nitems, FILE* file);
   void mfread(float* dummy, const int nbytes, const int nitems, FILE* file);

   double* fill_disttable(const char* const distfile);

   int precision(double number);

   float ibm2ieee(const float ibm);
   
   double ibm2ieee(const double ibm);

   if (argc < 7 || argc > 13) {
      printf("\n");
      printf("%s converts a single shot from a standard (IBM floating point) or\n",argv[0]);
      printf("non-standard (IEEE floating point) SEG-Y file\n");
      printf("to ascii using x,y,z format where x is the chan, cdp, or distance,\n");
      printf("y is the time or depth, and z is the amplitude of the trace\n");
      printf("\n");
      printf("%s will place only the given shot# in the output file,\n",argv[0]);
      printf("   where the shot# argument is the incremental shot number, not FFID,\n");
      printf("   and the SEGY file must have a constant number of traces per shot\n");
      printf("\n");
      printf("%s will place each trace from the shot sequentially in the output file,\n",argv[0]);
      printf("separated by a line containing \"#\"\n");
      printf("\n");
      printf("%s will read the chan, cdp, sample rate, and\n",argv[0]);
      printf("number of samples directly from the SEG-Y file\n");
      printf("\n");
      printf("%s will shift the traces' y values by datum amount, where datum\n",argv[0]);
      printf("is in terms of time or depth (positive datum shifts traces downward)\n");
      printf("\n");
      printf("mode must be \"chan\" or \"cdp\" or \"dist\"\n");
      printf("\n");
      printf("if mode is \"dist\", then %s will convert x values from cdp to distance,\n",argv[0]);
      printf("   and the CDPX_filename argument MUST be included\n");
      printf("   CDPX filename format:  cdp#  distance\n");
      printf("                example:     1      0.31\n");
      printf("                             2     10.20\n");
      printf("                           ...     .....\n");
      printf("note:  every cdp in the SEG-Y file must have a\n");
      printf("       corresponding distance value in the CDPX file\n");
      printf("\n");
      printf("%s will optionally limit the output to specified\n",argv[0]);
      printf("x_start, x_end, y_start, and y_end values if these arguments are given\n");
      printf("\n");
      printf("use \"-IEEE\" as last argument if SEGY-Y file has sample values in 4 byte IEEE format\n");
      printf("(default is 4 byte IBM floating point format)\n");
      printf("\n");
      printf("format:  %s  input_segy_file  output_ASCII_file  shot#  traces_per_shot  datum  mode\n",
         argv[0]);
      printf("  [CDPX_filename]  [x_start  x_end  y_start  y_end]  [-IEEE]\n");
      printf("\n");
      printf("example: %s  segy_file  output  1  60    0  chan\n",argv[0]);
      printf("example: %s  segy_file  output  1  60  -12  dist  CDPX\n",argv[0]);
      printf("example: %s  segy_file  output  1  60    0   cdp        10  3000  0  100\n",argv[0]);
      printf("example: %s  segy_file  output  1  60  -18  dist  CDPX  10  3000  0  100  -ieee\n",
         argv[0]);
      printf("\n");
      exit(1);
   }

   FILE* in;
   FILE* out;

   char dummy[3601];  // dummy buffer for skipping unused header information

   short numsamp;
   short samprate;
   int chan;
   int cdp;
   int xmode;         // mode code:  0 = chan, 1 = cdp, 2 = dist
   int x_start = 0;   // used if limitflag = 1
   int x_end   = 0;   // used if limitflag = 1
   int y_start = 0;   // used if limitflag = 1
   int y_end   = 0;   // used if limitflag = 1
   int limitflag = 0; // limit flag:   0 = use all data, 1 = limit data
   double time;

   // GET MANDATORY ARGUMENTS
   const char* input = argv[1];
   const char* output = argv[2];

   int shotNum = atoi(argv[3]);
   if (shotNum < 1) {
      printf("\nERROR:  shot# argument < 1\n\n");
      exit(1);
   }
   int tracesPerShot = atoi(argv[4]);
   if (tracesPerShot < 1) {
      printf("\nERROR:  traces_per_shot argument < 1\n\n");
      exit(1);
   }
   int datum = atoi(argv[5]);
   const char* mode = argv[6];

   // CHECK IF LAST ARGUMENT IS "-IEEE" or "IEEE"
   int ieee = 0;
   if (strcasecmp(argv[argc-1], "-ieee") == 0) {
      ieee = 1;
      argc--;
   }
   else if (strcasecmp(argv[argc-1], "ieee") == 0) {
      ieee = 1;
      argc--;
   }

   // SET XMODE PARAMETER
   if (strcasecmp(mode,"chan") == 0)
      xmode = 0;
   else if (strcasecmp(mode,"cdp") == 0)
      xmode = 1;
   else if (strcasecmp(mode,"dist") == 0)
      xmode = 2;
   else {
      printf("\nERROR:  mode is %s, must be \"chan\", \"cdp\", or \"dist\"\n\n",mode);
      exit(1);
   }

   // IF MODE IS DIST, THEN GET DISTFILE ARGUMENT
   char distfile[200];
   strcpy(distfile, "");
   double* disttable = 0;
   int distflag = 0;
   if (xmode == 2) {
      if (argc < 8) {
         printf("\nERROR:  mode is \"dist\" but no CDPX_filename argument given\n\n",mode);
         exit(1);
      }
      strcpy(distfile,argv[7]);
      disttable = fill_disttable(distfile);
      distflag = 1;
   }

   // CHECK NUMBER OF ARGUMENTS
   if (distflag == 0) {
      if (argc != 7 && argc != 11) {
         printf("\nERROR:  cannot have CDPX_filename argument unless mode is \"dist\"\n\n");
         exit(1);
      }
   }
   else {
      if (argc != 8 && argc != 12) {
         printf("\nERROR:  must have CDPX_filename argument if mode is \"dist\"\n\n");
         exit(1);
      }
   }

   // IF VALID, SET TIME PARAMETERS
   if ( argc > (7 + distflag) ) {
      if ( argc != (11 + distflag) ) {
         printf("\nERROR:  wrong number of arguments\n\n");
         exit(1);
      }
      limitflag = 1;
      int nextarg = 7 + distflag;
      x_start = atoi(argv[nextarg]);
      x_end   = atoi(argv[nextarg + 1]);
      y_start = atoi(argv[nextarg + 2]);
      y_end   = atoi(argv[nextarg + 3]);
      if (x_end < x_start) {
         printf("\nERROR:  x_end value of %d is less than x_start value of %d\n\n",x_end,x_start);
         exit(1);
      }
      if (y_end < y_start) {
         printf("\nERROR:  y_end value of %d is less than y_start value of %d\n\n",y_end,y_start);
         exit(1);
      }
   }

   // OPEN INPUT AND OUTPUT FILES
   if ((in = fopen(input,"r")) == NULL) {
      printf("\nERROR:  cannot find file %s\n\n",input);
      exit(1);
   }
   if ((out = fopen(output,"w")) == NULL) {
      printf("\nERROR:  cannot open file %s\n\n",output);
      exit(1);
   }

   // GET NUMBER_SAMPLES, SAMPLE_RATE FROM REEL HEADER INFORMATION
   // (SKIP TO CORRECT LOCATION IN HEADER FILE BY
   // READING IN UNUSED BYTES TO A DUMMY VARIABLE)
   mcread(dummy,     3216, 1, in);
   msread(&samprate,    2, 1, in);
   mcread(dummy,        2, 1, in);
   msread(&numsamp,     2, 1, in);
   mcread(dummy,      378, 1, in);
   double fsamprate = samprate / 1000.0;
   int prec = precision(fsamprate);

   // READ TRACES, INCLUDING CHAN AND CDP

   float value;

   int printflag;

   int poundflag;

   int currentShot;
   int previousShot = -1;

   int shotCounter = 1;
   int traceCounter = 1;

//////////////////////////////////////////////////////////////
//
// START MAIN LOOP
//
//////////////////////////////////////////////////////////////

   while (shotCounter <= shotNum && traceCounter <= tracesPerShot) {

      mcread (dummy,    8, 1, in);

      miread (&currentShot,  4, 1, in);

      miread (&chan,  4, 1, in);

      if (previousShot == -1) {
         previousShot = currentShot;
      }

      if (previousShot != currentShot) {
         if (traceCounter != 1) {
            printf("\nERROR:  only %d traces in SEGY file for shot %d\n\n", traceCounter, shotCounter);
            exit(1);
         }
         else {
            previousShot = currentShot;
         }
      }

      mcread(dummy,   4, 1, in);
      miread(&cdp,    4, 1, in);
      mcread(dummy, 216, 1, in);

      poundflag = 0;

      for (int i = 0 ; i < numsamp; i++) {

         mfread(&value, 4, 1, in);

         time = (double)i * fsamprate + (double)datum;

         printflag = 1;

         if (limitflag == 1) {
            if (xmode == 0) {
               if (x_start > chan) printflag = 0;
               if (x_end   < chan) printflag = 0;
            }
            if (xmode == 1) {
               if (x_start > cdp) printflag = 0;
               if (x_end   < cdp) printflag = 0;
            }
            if (xmode == 2) {
               if ( (double)x_start > disttable[cdp] ) printflag = 0;
               if ( (double)x_end   < disttable[cdp] ) printflag = 0;
            }
            if ( (double)y_start > time ) printflag = 0;
            if ( (double)y_end   < time ) printflag = 0;
         }

         if ( printflag && (shotCounter == shotNum) ) {

            poundflag++;
            if (poundflag == 1)
               fprintf(out,"#\n");

            if (ieee != 1)
               value = ibm2ieee(value);

            if      (xmode == 0)
               fprintf(out,"%4d   %8.*f %12.4f\n",chan,           prec, time, value);
            else if (xmode == 1)
               fprintf(out,"%4d   %8.*f %12.4f\n",cdp,            prec, time, value);
            else if (xmode == 2)
               fprintf(out,"%8.2f %8.*f %12.4f\n",disttable[cdp], prec, time, value);
         }

      }

      traceCounter++;
      if (traceCounter > tracesPerShot) {
         traceCounter = 1;
         shotCounter++;
         if (shotCounter > shotNum)
            break;
      }

   }

   if (shotCounter < shotNum) {
      if (shotCounter == 1)
         printf("\nERROR:  only %d shot found in SEGY file\n\n", shotCounter);
      else
         printf("\nERROR:  only %d shots found in SEGY file\n\n", shotCounter);
      exit(1);
   }

   fclose(in);
   fclose(out);
   if (xmode == 2)
      delete[] disttable;

}

//////////////////////////////////////////////////////////////
//
// SUBROUTINES
//
//////////////////////////////////////////////////////////////

void mcread(char* dummy, const int nbytes, const int nitems, FILE* file) {
   if (fread (dummy, nbytes, nitems, file) == 0) {
      printf("\nERROR:  unexpected end of SEGY file\n");
      exit(1);
   }
}
void msread(short* dummy, const int nbytes, const int nitems, FILE* file) {
   if (fread (dummy, nbytes, nitems, file) == 0) {
      printf("\nERROR:  unexpected end of SEGY file\n");
      exit(1);
   }
}
void miread(int* dummy, const int nbytes, const int nitems, FILE* file) {
   if (fread (dummy, nbytes, nitems, file) == 0) {
      printf("\nERROR:  unexpected end of SEGY file\n");
      exit(1);
   }
}
void mfread(float* dummy, const int nbytes, const int nitems, FILE* file) {
   if (fread (dummy, nbytes, nitems, file) == 0) {
      printf("\nERROR:  unexpected end of SEGY file\n");
      exit(1);
   }
}

double* fill_disttable(const char* const distfile) {
   
   // CDPX FILE MAY BE IN FREE-FORM FORMAT OR IN PROMAX FORMAT

   FILE* ft;
   char string[200];
   int trace;
   int maxtrace = 0;
   double value;

   if ((ft = fopen(distfile, "r")) == NULL) {
      printf("\nERROR:  cannot open file %s\n\n",distfile);
      exit(1);
   }

   while (fgets(string,199,ft) != NULL) {
      if (string[0] != '<') {
         if (sscanf(string,"%d %lf",&trace, &value) != 2)
            continue;
      }
      else {
         if (sscanf(string,"%*c %d %*c %lf",&trace, &value) != 2) {
            printf("\nERROR:  bad line in CDPX file %s\n\n", distfile);
            exit(1);
         }
      }
      if (maxtrace < trace)
         maxtrace = trace;
   }

   rewind(ft);

   double* disttable = new double[maxtrace + 1];
   for (int i = 0; i < (maxtrace + 1); i++)
      disttable[i] = 0.0;

   while (fgets(string,199,ft) != NULL) {
      if (string[0] != '<') {
         if (sscanf(string,"%d %lf",&trace, &value) != 2)
            continue;
      }
      else {
         if (sscanf(string,"%*c %d %*c %lf",&trace, &value) != 2) {
            printf("\nERROR:  bad line in CDPX file %s\n\n", distfile);
            exit(1);
         }
      }
      disttable[trace] = value;
   }

   fclose(ft);
   return disttable;

}

int precision(double number) {

   int dec = 0;
   while (fabs(number - rint(number)) >= .00001) {
      number *= 10.0;
      dec++;
      if (dec >= 7) 
         break;
   }

   return dec;
   
}

float ibm2ieee(const float ibm) {

   // DEFINE AND SETUP UNION VARIABLES
   union value {
      float f;
      unsigned char c[4];
   };
   value src;

   // ASSIGN PARAMETER TO UNION
   src.f = ibm;

   // CONVERT TO FLOAT
   long IntMantissa =
        ((long)src.c[1] << 16)
      + ((long)src.c[2] << 8)
      + src.c[3];

   float  Mantissa  = float(IntMantissa) / float(0x1000000);
   float  PosResult = Mantissa * pow( 16.0, double((src.c[0] & 0x7F) - 64) );

   if (src.c[0] & 0x80)
      return -PosResult;
   else
      return PosResult;

}

double ibm2ieee(const double ibm) {

   // DEFINE AND SETUP UNION VARIABLES
   union value {
      double f;
      unsigned char c[8];
   };
   value src;

   // ASSIGN PARAMETER TO UNION
   src.f = ibm;

   // CONVERT TO FLOAT
   long long IntMantissa =
        ((long long)src.c[1] << 48)
      + ((long long)src.c[2] << 40)
      + ((long long)src.c[3] << 32)
      + ((long long)src.c[4] << 24)
      + ((long long)src.c[5] << 16)
      + ((long long)src.c[6] << 8)
      + src.c[7];

   double  Mantissa  = double(IntMantissa) / double(0x1000000);
   double  PosResult = Mantissa * pow( 16.0, double((src.c[0] & 0x7F) - 64) );

   if (src.c[0] & 0x80)
      return -PosResult;
   else
      return PosResult;

}

