#include<cstdio>
#include<cstdlib>
#include<cstring>
#include<cmath>

///////////////////////////////////////
//
// WRITTEN BY MARK GOLDMAN
// U.S. GEOLOGICAL SURVEY
// MS 977
// 345 MIDDLEFIELD RD
// MENLO PARK, CALIFORNIA 94025
// goldman@usgs.gov
//
///////////////////////////////////////

 void readSegy(int argc, char** argv);
float ibm2ieee(const float ibm);
double ibm2ieee(const double ibm);

main(int argc, char** argv) {
   readSegy(argc, argv);
}

void readSegy(int argc, char** argv) {

   if (argc != 3 && argc != 4 && argc != 5 && argc != 6) {
      printf("\n");
      printf("%s reads a SEGY file (having a 3600 byte reel header),\n",argv[0]);
      printf("displays header info and min/max sample info,\n");
      printf("and optionally converts the sample values to ASCII\n");
      printf("\n");
      printf("input file trace sample format code is:\n");
      printf("   1 = 4 byte IBM floating point\n");
      printf("   2 = 4 byte integer\n");
      printf("   3 = 2 byte integer\n");
      printf("   5 = 4 byte IEEE floating point\n");
      printf("\n");
      printf("maximum 32767 samples per trace, read from bytes 115-116\n");
      printf("\n");
      printf("optional #_traces, samples/trace arguments overide header values\n");
      printf("\n");
      printf("format:  %s  input_segy_filename  format_code  [output_ascii_file]",argv[0]);
      printf("  [#_traces  samples/trace]\n");
      printf("\n");
      printf("example: %s  input  1\n",argv[0]);
      printf("example: %s  input  1  zout\n",argv[0]);
      printf("example: %s  input  1  60  2000\n",argv[0]);
      printf("example: %s  input  1  zout  60  2000\n",argv[0]);
      printf("\n");
      exit(1);
   }

   FILE* in;
   FILE* out;

   char dummy[3601];  // dummy buffer for skipping unused header information

   char  c[200];
   short s;
   int   i;
   float f;

   char input[200];
   char output[200];

   strcpy(input, argv[1]);

   int code = atoi(argv[2]);
   if (code != 1 && code != 2 && code != 3 && code != 5) {
      printf("\nERROR:  input_format argument must be 1, 2, 3 or 5\n\n");
      exit(1);
   }

   int convert;
   int numtraces = 0;
   int numsamptrue = 0;

   if (argc == 4 || argc == 6) {
      convert = 1;
      strcpy(output,argv[3]);
   }
   else {
      convert = 0;
   }

   if (argc == 5 || argc == 6) {
      numtraces   = atoi(argv[argc - 2]);
      numsamptrue = atoi(argv[argc - 1]);
      if (numtraces < 1) {
         printf("\nERROR:  #_traces argument must be > 0\n\n");
         exit(1);
      }
      if (numsamptrue < 1) {
         printf("\nERROR:  samples/trace argument must be > 0\n\n");
         exit(1);
      }
   }

   // OPEN INPUT AND OUTPUT FILES
   if ((in = fopen(input,"r")) == NULL) {
      printf("\nERROR:  cannot find input file %s\n\n",input);
      exit(1);
   }

   if (convert) {
      if ((out = fopen(output,"w")) == NULL) {
         printf("\nERROR:  cannot open output file %s\n\n",output);
         exit(1);
      }
   }

/////////////////////////////////////////////////////

   // GET BINARY HEADER INFO

   printf("\nFOR BINARY REEL HEADER:\n");

   fread (&dummy,    3200, 1, in);

   fread (&dummy,       4, 1, in);

   fread (&i,           4, 1, in);
   printf("%6d = line number\n", i);

   fread (&dummy,       4, 1, in);

   fread (&s,           2, 1, in);
   printf("%6d = # data traces\n", s);

   fread (&s,           2, 1, in);
   printf("%6d = # aux traces\n", s);

   fread (&s,           2, 1, in);
   printf("%6d = reel sample interval in us (microseconds)\n", s);

   fread (&s,           2, 1, in);
   printf("%6d = original sample interval in us (microseconds)\n", s);

   fread (&s,           2, 1, in);
   printf("%6d = reel samples per trace\n", s);

   fread (&s,           2, 1, in);
   printf("%6d = original samples per trace\n", s);

   fread (&s,           2, 1, in);
   printf("%6d = data format code (1=IBM, 2=4byte_int, 3=2byte_int, 5=IEEE)\n", s);

   fread (&dummy,     374, 1, in);

///////////////////////////////////////////////////

   ////////////////////////////////////
   // READ TRACES
   ////////////////////////////////////

   int runflag = 1;
   int trace = 0;
   int ffid;
   int chan;
   int numsamp;

   short svalue;

   int ivalue;
   int imaxvalue;
   int imaxsamp;
   int iminvalue;
   int iminsamp;

   float fvalue;
   float fmaxvalue;
     int fmaxsamp;
   float fminvalue;
     int fminsamp;

   while (runflag) {

      //////////////////////////////////////////////////////
      // GET NUMBER_SAMPLES, SAMPLE_RATE FROM HEADER INFORMATION
      // (SKIP TO CORRECT LOCATION IN HEADER FILE BY
      // READING IN UNUSED BYTES TO A DUMMY VARIABLE)
      //////////////////////////////////////////////////////


////////////////////  STANDARD HEADER BYTES 1-180  ///////////////


      if (fread (&dummy,       8, 1, in) != 1) {
         runflag = 0;
         continue;
      }

      trace++;
      printf("\nFOR TRACE %d\n",trace);

      fread (&i,           4, 1, in);
      printf("%6d = ffid\n", i);
      ffid = i;

      fread (&i,           4, 1, in);
      printf("%6d = trace number\n", i);
      chan = i;

      fread (&dummy,       4, 1, in);

      fread (&i,           4, 1, in);
      printf("%6d = cdp number\n", i);

      fread (&dummy,       4, 1, in);

      fread (&s,           2, 1, in);
      printf("%6d = trace id code (1 = seismic, 2 = dead, 3 = dummy)\n", s);
   
      fread (&dummy,       6, 1, in);

      fread (&i,           4, 1, in);
      printf("%6d = offset\n", i);

      fread (&i,           4, 1, in);
      printf("%6d = receiver elevation\n", i);

      fread (&i,           4, 1, in);
      printf("%6d = source elevation\n", i);

      fread (&i,           4, 1, in);
      printf("%6d = source depth\n", i);

      fread (&dummy,      16, 1, in);

      fread (&s,           2, 1, in);
      printf("%6d = scalar for above elevations (+ multiplication, - division)\n", s);

      fread (&s,           2, 1, in);
      printf("%6d = scalar for below coordinates (+ multiplication, - division)\n", s);

      fread (&i,           4, 1, in);
      printf("%6d = source x coordinate OR seconds of longitude (pos = E)\n", i);

      fread (&i,           4, 1, in);
      printf("%6d = source y coordinate OR seconds of latitude (pos = N)\n", i);

      fread (&i,           4, 1, in);
      printf("%6d = receiver x coordinate OR seconds of longitude (pos = E)\n", i);

      fread (&i,           4, 1, in);
      printf("%6d = receiver y coordinate OR seconds of latitude (pos = N)\n", i);

      fread (&s,           2, 1, in);
      printf("%6d = coordinate units (1 = length, 2 = seconds of lat/lon)\n", s);

      fread (&dummy,      24, 1, in);

      fread (&s,           2, 1, in);
      printf("%6d = # samples (max 32767)\n", s);
      numsamp = (int)s;
      if (numsamptrue > 0)
         numsamp = numsamptrue;
   
      fread (&s,           2, 1, in);
      printf("%6d = sample interval in us (microseconds)\n", s);
   
      fread (&s,           2, 1, in);
      printf("%6d = gain type (1 = fixed, 2 = binary, 3 = floating point)\n", s);
   
      fread (&s,           2, 1, in);
      printf("%6d = instrument gain constant\n", s);
   
      fread (&s,           2, 1, in);
      printf("%6d = instrument initial gain\n", s);

      fread (&dummy,      32, 1, in);

      fread (&s,           2, 1, in);
      printf("%6d = year data recorded\n", s);

      fread (&s,           2, 1, in);
      printf("%6d = day data recorded\n", s);

      fread (&s,           2, 1, in);
      printf("%6d = hour data recorded\n", s);

      fread (&s,           2, 1, in);
      printf("%6d = minute data recorded\n", s);

      fread (&s,           2, 1, in);
      printf("%6d = second data recorded\n", s);

      fread (&s,           2, 1, in);
      printf("%6d = time basis code (1 = local, 2 = GMT, 3 = other)\n", s);

      fread (&dummy,       6, 1, in);

      fread (&s,           2, 1, in);
      printf("%6d = DAS instrument # plus 10,000 (valid if .SUX file)\n", s);

      fread (&dummy,       4, 1, in);

//////////////////////////  EXTENDED HEADER BYTES 181-240  ///////////////

      fread (&dummy,      60, 1, in);

      //////////////////////////////////////////////////////
      // GET SAMPLE VALUES
      //////////////////////////////////////////////////////

      if (convert)
         fprintf(out, "# ffid,chan,trace:  %6d  %6d  %6d\n", ffid, chan, trace);

      if (code == 1 || code == 5) {

         fmaxvalue  = 0.0;
         fmaxsamp   = 0;
         fminvalue  = 0.0;
         fminsamp   = 0;
   
         for (int i = 0 ; i < numsamp; i++) {
   
            if (fread(&fvalue, 4, 1, in) != 1) {
               printf("\nERROR:  unexpected end of segy file %s\n\n",input);
               exit(1);
            }
   
            if (code == 1) 
               fvalue = ibm2ieee(fvalue);

            if (convert)
               fprintf(out, "%6d  %20.10f\n", i+1, fvalue);
   
            if (fvalue > fmaxvalue) {
               fmaxvalue = fvalue;
               fmaxsamp = i+1;
            }
            if (fvalue < fminvalue) {
               fminvalue = fvalue;
               fminsamp = i+1;
            }

         }

         printf("max at samp, min at samp: %20.10f at %6d, %20.10f at %6d\n", 
            fmaxvalue, fmaxsamp, fminvalue, fminsamp);

      }

      if (code == 2 || code == 3) {

         imaxvalue  = 0;
         imaxsamp   = 0;
         iminvalue  = 0;
         iminsamp   = 0;
   
         for (int i = 0 ; i < numsamp; i++) {
   
            if (code == 2) {
               if (fread(&ivalue, 4, 1, in) != 1) {
                  printf("\nERROR:  unexpected end of segy file %s\n\n",input);
                  exit(1);
               }
            }
            if (code == 3) {
               if (fread(&svalue, 2, 1, in) != 1) {
                  printf("\nERROR:  unexpected end of segy file %s\n\n",input);
                  exit(1);
               }
               ivalue = (int)svalue;
            }
   
            if (convert)
               fprintf(out, "%6d  %20d\n", i+1, ivalue);
   
            if (ivalue > imaxvalue) {
               imaxvalue = ivalue;
               imaxsamp = i+1;
            }
            if (ivalue < iminvalue) {
               iminvalue = ivalue;
               iminsamp = i+1;
            }
   
         }

         printf("max at samp, min at samp: %20d at %6d, %20d at %6d\n", 
            imaxvalue, imaxsamp, iminvalue, iminsamp);

      }

      printf("\n");

   }

   fclose(in);

   if (convert)
      fclose(out);

   if ( (numtraces > 0) && (trace != numtraces) ) {
      printf("\nWARNING:  #_traces given as %d, but %d traces found\n\n", numtraces, trace);

   }
}

//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////

float ibm2ieee(const float ibm) {

   // DEFINE AND SETUP UNION VARIABLES
   union value {
      float f;
      unsigned char c[4];
   };
   value src;

   // ASSIGN PARAMETER TO UNION
   src.f = ibm;

   // CONVERT TO FLOAT
   long IntMantissa =
        ((long)src.c[1] << 16)
      + ((long)src.c[2] << 8)
      + src.c[3];

   float  Mantissa  = float(IntMantissa) / float(0x1000000);
   float  PosResult = Mantissa * pow( 16.0, double((src.c[0] & 0x7F) - 64) );

   if (src.c[0] & 0x80)
      return -PosResult;
   else
      return PosResult;

}

double ibm2ieee(const double ibm) {

   // DEFINE AND SETUP UNION VARIABLES
   union value {
      double f;
      unsigned char c[8];
   };
   value src;

   // ASSIGN PARAMETER TO UNION
   src.f = ibm;

   // CONVERT TO FLOAT
   long long IntMantissa =
        ((long long)src.c[1] << 48)
      + ((long long)src.c[2] << 40)
      + ((long long)src.c[3] << 32)
      + ((long long)src.c[4] << 24)
      + ((long long)src.c[5] << 16)
      + ((long long)src.c[6] << 8)
      + src.c[7];

   double  Mantissa  = double(IntMantissa) / double(0x1000000);
   double  PosResult = Mantissa * pow( 16.0, double((src.c[0] & 0x7F) - 64) );

   if (src.c[0] & 0x80)
      return -PosResult;
   else
      return PosResult;

}

