#include<cstdio>
#include<cstdlib>
#include<cstring>
#include<cmath>

///////////////////////////////////////
//
// WRITTEN BY MARK GOLDMAN
// U.S. GEOLOGICAL SURVEY
// MS 977
// 345 MIDDLEFIELD RD
// MENLO PARK, CALIFORNIA 94025
// goldman@usgs.gov
//
///////////////////////////////////////

main(int argc, char** argv) {
 
   int precision(double number);

   float ibm2ieee(const float ibm);
   
   double ibm2ieee(const double ibm);

   void convert2(char* input);
   void convert4(char* input);

   float vfloat(float vax);

   float efloat(float value);

   if (argc != 4) {
      printf("\n");
      printf("%s byte-swaps SEGY file header values between little-endian and big-endian,\n",argv[0]);
      printf("   optionally changing data values as well\n");
      printf("\n");
      printf("%s will read the chan, cdp, sample rate, and\n",argv[0]);
      printf("number of samples directly from the SEGY file\n");
      printf("\n");
      printf("format:  %s  input_file  output_file  data_conversion_code\n",argv[0]);
      printf("\n");
      printf("data_conversion_code (affects only sample values, not header values):\n");
      printf("   1 = don't convert data\n");
      printf("       (data in IBM floating point format does not need byte-swapping)\n");
      printf("   2 = convert data from IBM floating point to IEEE floating point\n");
      printf("   3 = convert data from IEEE floating point to IEEE floating point\n");
      printf("       (byte-swap data)\n");
      printf("   4 = convert data from VAX floating point to IEEE floating point\n");
      printf("\n");
      printf("note:  if data is in IBM floating point format, then use option 1 or 2\n");
      printf("       if data is in IEEE floating point format, then use option 3\n");
      printf("       if data is in VAX floating point format, then use option 4\n");
      printf("\n");
      exit(1);
   }

   FILE* in;
   FILE* out;

   union {
      char  c[2];
      short s;
   } u2;

   union {
      char  c[4];
      int   i;
      float f;
   } u4;

   short numsamp;
   short samprate;
   short code;
   short measure;
   int   vred;

   int ffid;
   int chan;
   int offset;
   short elev_scaler;
   short coord_scaler;
   short tnumsamp;
   short tsampint;

   char dummy[3601];  // dummy buffer for skipping unused header information

   // GET MANDATORY ARGUMENTS
   const char* input = argv[1];
   const char* output = argv[2];

   int convertFlag = atoi(argv[3]);
   if (convertFlag < 1 || convertFlag > 4) {
      printf("\nERROR:  data conversion code must be between 1 and 4\n\n");
      exit(1);
   }

   // OPEN INPUT AND OUTPUT FILES
   if ((in = fopen(input,"r")) == NULL) {
      printf("\nERROR:  cannot find file %s\n\n",input);
      exit(1);
   }
   if ((out = fopen(output,"w")) == NULL) {
      printf("\nERROR:  cannot open file %s\n\n",output);
      exit(1);
   }

///////////////////////////////////
//
// REEL HEADER
//
///////////////////////////////////

   // GET NUMBER_SAMPLES, SAMPLE_RATE FROM HEADER INFORMATION
   // (SKIP TO CORRECT LOCATION IN HEADER FILE BY
   // READING IN UNUSED BYTES TO A DUMMY VARIABLE)
   fread(&dummy,   3200, 1, in);
   fwrite(&dummy,  3200, 1, out);

   // JOB ID
   fread(&dummy,      4, 1, in);
   convert4(dummy);
   fwrite(&dummy,     4, 1, out);

   // LINE #
   fread(&dummy,      4, 1, in);
   convert4(dummy);
   fwrite(&dummy,     4, 1, out);

   // REEL #
   fread(&dummy,      4, 1, in);
   convert4(dummy);
   fwrite(&dummy,     4, 1, out);

   // NUM DATA
   fread(&dummy,      2, 1, in);
   convert2(dummy);
   fwrite(&dummy,     2, 1, out);

   // NUM AUX 
   fread(&dummy,      2, 1, in);
   convert2(dummy);
   fwrite(&dummy,     2, 1, out);

   // SAMPLE INT IN MICROSECONDS, THIS REEL
   fread(&dummy,      2, 1, in);
   convert2(dummy);
   fwrite(&dummy,     2, 1, out);
   u2.c[0] = dummy[0];
   u2.c[1] = dummy[1];
   samprate = u2.s;

   // SAMPLE INT IN MICROSECONDS, ORIGINAL
   fread(&dummy,      2, 1, in);
   convert2(dummy);
   fwrite(&dummy,     2, 1, out);

   // NUMBER SAMPLES, THIS REEL
   fread(&dummy,      2, 1, in);
   convert2(dummy);
   fwrite(&dummy,     2, 1, out);
   u2.c[0] = dummy[0];
   u2.c[1] = dummy[1];
   numsamp = u2.s;

   // NUMBER SAMPLES, ORIGINAL
   fread(&dummy,      2, 1, in);
   convert2(dummy);
   fwrite(&dummy,     2, 1, out);

   // DATA SAMPLE FORMAT 
   // 1 = IBM floating point
   // 2 = 4 byte fixed point 
   // 3 = 2 byte fixed point 
   // 4 = 4 bytes fixed point with gain
   // 5 = IEEE floating point
   //
   fread(&dummy,      2, 1, in);
   convert2(dummy);
   u2.c[0] = dummy[0];
   u2.c[1] = dummy[1];
   code = u2.s;
   if (convertFlag == 2 || convertFlag == 4) {
      code = 5;
      u2.s = 5;
      dummy[0] = u2.c[0];
      dummy[1] = u2.c[1];
   }
   fwrite(&dummy,     2, 1, out);

   fread(&dummy,     28, 1, in);
   fwrite(&dummy,    28, 1, out);
   
   // MEASUREMENT SYSTEM (1 = meters, 2 = feet)
   fread(&dummy,      2, 1, in);
   convert2(dummy);
   fwrite(&dummy,     2, 1, out);
   u2.c[0] = dummy[0];
   u2.c[1] = dummy[1];
   measure = u2.s;

   fread(&dummy,     16, 1, in);
   fwrite(&dummy,    16, 1, out);
   
   // REDUCTION VELOCITY
   fread(&dummy,      4, 1, in);
   convert4(dummy);
   fwrite(&dummy,     4, 1, out);
   u4.c[0] = dummy[0];
   u4.c[1] = dummy[1];
   u4.c[2] = dummy[2];
   u4.c[3] = dummy[3];
   vred = u4.i;

   fread(&dummy,    324, 1, in);
   fwrite(&dummy,   324, 1, out);
   
   // READ TRACES, INCLUDING CHAN AND CDP
   float fsamprate = samprate / 1000.0;
   int prec = precision(fsamprate);

   printf("\n");
   printf("REEL HEADER VALUES\n");
   printf("samprate is %6d  microseconds\n",samprate);
   printf("samprate is %6.*f  milliseconds\n",prec, fsamprate);
   printf("numsamp  is %6d  samples per trace\n",numsamp);
   printf("data sample format code is %d\n",code);
   printf("measurement system code is %d\n",measure);
   printf("reduction velocity is %6d\n",vred);
   printf("\n");

   float value;
   int trace = 0;

///////////////////////////////////
//
// TRACE HEADER
//
///////////////////////////////////

   // TRACE #
   while (fread(&dummy,    4, 1, in) == 1) {
      convert4(dummy);
      fwrite(&dummy,    4, 1, out);

      // TRACE #
      fread(&dummy, 4, 1, in);
      convert4(dummy);
      fwrite(&dummy, 4, 1, out);

      // FFID
      fread(&dummy, 4, 1, in);
      convert4(dummy);
      fwrite(&dummy, 4, 1, out);
      u4.c[0] = dummy[0];
      u4.c[1] = dummy[1];
      u4.c[2] = dummy[2];
      u4.c[3] = dummy[3];
      ffid = u4.i;

      // CHAN
      fread(&dummy, 4, 1, in);
      convert4(dummy);
      fwrite(&dummy, 4, 1, out);
      u4.c[0] = dummy[0];
      u4.c[1] = dummy[1];
      u4.c[2] = dummy[2];
      u4.c[3] = dummy[3];
      chan = u4.i;

      fread(&dummy,   4, 1, in);
      fwrite(&dummy,   4, 1, out);

      // CDP 
      fread(&dummy, 4, 1, in);
      convert4(dummy);
      fwrite(&dummy, 4, 1, out);

      // CDP TRACE NUMBER
      fread(&dummy, 4, 1, in);
      convert4(dummy);
      fwrite(&dummy, 4, 1, out);

      // TRACE ID CODE
      // 1 = data, 2 = dead, 3 = dummy, 4 = time break, 5 = uphole,
      // 6 = sweep, 7 = timing, 8 = water break, 9 = other
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      fread(&dummy,   6, 1, in);
      fwrite(&dummy,   6, 1, out);

      // OFFSET
      fread(&dummy, 4, 1, in);
      convert4(dummy);
      fwrite(&dummy, 4, 1, out);
      u4.c[0] = dummy[0];
      u4.c[1] = dummy[1];
      u4.c[2] = dummy[2];
      u4.c[3] = dummy[3];
      offset = u4.i;

      // RECEIVER ELEVATION
      fread(&dummy, 4, 1, in);
      convert4(dummy);
      fwrite(&dummy, 4, 1, out);

      // SOURCE ELEVATION
      fread(&dummy, 4, 1, in);
      convert4(dummy);
      fwrite(&dummy, 4, 1, out);

      // SOURCE DEPTH
      fread(&dummy, 4, 1, in);
      convert4(dummy);
      fwrite(&dummy, 4, 1, out);

      fread(&dummy,  16, 1, in);
      fwrite(&dummy,  16, 1, out);

      // SCALAR TO ELEVATIONS AND DEPTHS (+ multiplication, - division)
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);
      u2.c[0] = dummy[0];
      u2.c[1] = dummy[1];
      elev_scaler = u2.s;

      // SCALAR TO COORDINATES (+ multiplication, - division)
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);
      u2.c[0] = dummy[0];
      u2.c[1] = dummy[1];
      coord_scaler = u2.s;

      // SOURCE X COORDINATE
      fread(&dummy, 4, 1, in);
      convert4(dummy);
      fwrite(&dummy, 4, 1, out);

      // SOURCE Y COORDINATE
      fread(&dummy, 4, 1, in);
      convert4(dummy);
      fwrite(&dummy, 4, 1, out);

      // RECEIVER X COORDINATE
      fread(&dummy, 4, 1, in);
      convert4(dummy);
      fwrite(&dummy, 4, 1, out);

      // RECEIVER Y COORDINATE
      fread(&dummy, 4, 1, in);
      convert4(dummy);
      fwrite(&dummy, 4, 1, out);

      // COORDINATE MEASURE CODE (1 = length, 2 = seconds of lat/lon)
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);
      u2.c[0] = dummy[0];
      u2.c[1] = dummy[1];
      coord_scaler = u2.s;

      fread(&dummy, 18, 1, in);
      fwrite(&dummy, 18, 1, out);

      // DELAY RECORDING TIME
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      fread(&dummy, 4, 1, in);
      fwrite(&dummy, 4, 1, out);

      // NUMBER OF SAMPLES
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);
      u2.c[0] = dummy[0];
      u2.c[1] = dummy[1];
      tnumsamp = u2.s;

      // SAMPLE INTERVAL IN MICROSECONDS
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);
      u2.c[0] = dummy[0];
      u2.c[1] = dummy[1];
      tsampint = u2.s;

      // GAIN TYPE
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      // GAIN CONSTANT
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      // INSTRUMENT INITIAL GAIN
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      fread(&dummy, 32, 1, in);
      fwrite(&dummy, 32, 1, out);

      // YEAR
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      // DAY
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      // HOUR
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      // MINUTE
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      // SECOND
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      // TIME BASIS CODE (1 = local, 2 = GMT, 3 = other)
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      fread(&dummy, 6, 1, in);
      fwrite(&dummy, 6, 1, out);

      // DAS INSTRUMENT # + 10,000 (valid if .SUX file)
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      fread(&dummy, 4, 1, in);
      fwrite(&dummy, 4, 1, out);

/////////////////////////////////////////
//
// EXTENDED TRACE HEADER (NON-STANDARD)
//
/////////////////////////////////////////

      // MILLISECONDS
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      // SHOTPOINT LOCATION NUMBER
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      // RECORDING INSTRUMENT UNIT NUMBER
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      fread(&dummy, 4, 1, in);
      fwrite(&dummy, 4, 1, out);

      // DISTANCE WEIGHTING EXPONENT (hundreths)
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      // SHOT SEQUENCE
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      // SHOT SIZE (kg)
      fread(&dummy, 2, 1, in);
      convert2(dummy);
      fwrite(&dummy, 2, 1, out);

      // SHOT POINT - STATION AZIMUTH (second of arc)
      fread(&dummy, 4, 1, in);
      convert4(dummy);
      fwrite(&dummy, 4, 1, out);

      // TIME OF FIRST POINT MINUS SHOT TIME (msec)
      fread(&dummy, 4, 1, in);
      convert4(dummy);
      fwrite(&dummy, 4, 1, out);

      fread(&dummy, 36, 1, in);
      fwrite(&dummy, 36, 1, out);

///////////////////////////////////
//
// TRACE VALUES
//
///////////////////////////////////

      trace++;

      for (int i = 0 ; i < numsamp; i++) {

         if (fread(&value, 4, 1, in) != 1) {
            printf("\nERROR:  unexpected end of segy file %s\n\n",input);
            exit(1);
         }

         if (convertFlag == 1)
            ; // no change

         if (convertFlag == 2)
            value = ibm2ieee(value);

         if (convertFlag == 3)
            value = efloat(value);

         if (convertFlag == 4)
            value = vfloat(value);

         fwrite(&value, 4, 1, out);

         if (i == 0) {
            printf("trace, numsamp, samp_int, ffid, chan, offset,    first_sample:\n");
            printf("%5d  %7d  %8d  %4d  %4d %7d",
               trace, tnumsamp, tsampint, ffid, chan, offset);
            if (convertFlag == 1)
               printf(" %16.6f\n", ibm2ieee(value));
            else
               printf(" %16.6f\n", value);
            printf("\n");
         }

      }

   }

   fclose(in);
   fclose(out);

}

///////////////////////////////////
//
// SUBROUTINES
//
///////////////////////////////////

void convert2(char* input) {

   unsigned char temp;

   temp = input[1];
   input[1] = input[0];
   input[0] = temp;

}

void convert4(char* input) {

   unsigned char temp;

   temp = input[3];
   input[3] = input[0];
   input[0] = temp;

   temp = input[2];
   input[2] = input[1];
   input[1] = temp;

}

int precision(double number) {

   int dec = 0;
   while (fabs(number - rint(number)) >= .00001) {
      number *= 10.0;
      dec++;
      if (dec >= 7) 
         break;
   }

   return dec;
   
}

float ibm2ieee(const float ibm) {

   // DEFINE AND SETUP UNION VARIABLES
   union value {
      float f;
      unsigned char c[4];
   };
   value src;

   // ASSIGN PARAMETER TO UNION
   src.f = ibm;

   // CONVERT TO FLOAT
   long IntMantissa =
        ((long)src.c[1] << 16)
      + ((long)src.c[2] << 8)
      + src.c[3];

   float  Mantissa  = float(IntMantissa) / float(0x1000000);
   float  PosResult = Mantissa * pow( 16.0, double((src.c[0] & 0x7F) - 64) );

   if (src.c[0] & 0x80)
      return -PosResult;
   else
      return PosResult;

}

double ibm2ieee(const double ibm) {

   // DEFINE AND SETUP UNION VARIABLES
   union value {
      double f;
      unsigned char c[8];
   };
   value src;

   // ASSIGN PARAMETER TO UNION
   src.f = ibm;

   // CONVERT TO FLOAT
   long long IntMantissa =
        ((long long)src.c[1] << 48)
      + ((long long)src.c[2] << 40)
      + ((long long)src.c[3] << 32)
      + ((long long)src.c[4] << 24)
      + ((long long)src.c[5] << 16)
      + ((long long)src.c[6] << 8)
      + src.c[7];

   double  Mantissa  = double(IntMantissa) / double(0x1000000);
   double  PosResult = Mantissa * pow( 16.0, double((src.c[0] & 0x7F) - 64) );

   if (src.c[0] & 0x80)
      return -PosResult;
   else
      return PosResult;

}

// routine that converts the vax floating point numbers to
// RT/IEEE floating point format

float vfloat(float vax) {

   union {
      char bytes[4];
      float flt;
      struct {
         unsigned int sign:1;
         unsigned int exp:8;
         unsigned int mant:23;
      } f;
   } ieee;

   union {
      float flt;
      char bytes[4];
   } vaxx;

// swap bytes
   vaxx.flt = vax;
   ieee.bytes[0] = vaxx.bytes[1];
   ieee.bytes[1] = vaxx.bytes[0];
   ieee.bytes[2] = vaxx.bytes[3];
   ieee.bytes[3] = vaxx.bytes[2];

// adjust exponent
   if (ieee.f.exp < 2)
      ieee.flt = 0.0;  /* underflows */
   else 
      ieee.f.exp -= 2;

// return
   return(ieee.flt);

}

float efloat(float value) {

   union {
      char bytes[4];
      float flt;
   } ieee;

// swap bytes
   unsigned char temp;
   ieee.flt = value;

   temp = ieee.bytes[3];
   ieee.bytes[3] = ieee.bytes[0];
   ieee.bytes[0] = temp;

   temp = ieee.bytes[2];
   ieee.bytes[2] = ieee.bytes[1];
   ieee.bytes[1] = temp;

// return
   return(ieee.flt);

}

