#include<cstdio>
#include<cstdlib>
#include<cstring>
#include<cmath>

///////////////////////////////////////
//
// WRITTEN BY MARK GOLDMAN
// U.S. GEOLOGICAL SURVEY
// MS 977
// 345 MIDDLEFIELD RD
// MENLO PARK, CALIFORNIA 94025
// goldman@usgs.gov
//
///////////////////////////////////////

main(int argc, char** argv) {
 
   if (argc != 6) {
      printf("\n");
      printf("%s converts from x,y,z format where x is the chan,\n", argv[0]);
      printf("y is the time or depth, and z is the amplitude of the trace\n");
      printf("to a non-standard (IEEE floating point) SEG-Y file\n");
      printf("\n");
      printf("traces may optionally be separated by one or more lines NOT in a format of x,y,z\n");
      printf("\n");
      printf("Example xyz input file for 3 traces, 3 samples per traces, sample interval 0.5 ms\n");
      printf("#\n");
      printf("1  0.0  -2.33\n");
      printf("1  0.5   1.03\n");
      printf("1  1.0  -4.03\n");
      printf("# comment: trace 2\n");
      printf("2  0.0  -3.33\n");
      printf("2  0.5   0.3\n");
      printf("2  1.0  -3.2\n");
      printf("3  0.0    .03\n");
      printf("3  0.5   -.34\n");
      printf("3  1.0   -.54\n");
      printf("\n");
      printf("Note:  SEGY format (and this program) assumes all traces start at time 0\n");
      printf("\n");
      printf("Note:  %s assumes all traces belong to FFID 1001\n",argv[0]);
      printf("\n");
      printf("format:  %s  input_xyz_file  output_SEGY_file\n   number_traces  number_samples_per_trace  sample_interval(ms)",argv[0]);
      printf("\n");
      printf("example: %s  xyz_file  segy_file  60  1024  2.0\n",argv[0]);
      printf("\n");
      exit(1);
   }

   void writeheader(int trace, int numsamples, double sampint, FILE* out);

   FILE* in;
   FILE* out;

   int numtraces = atoi(argv[3]);
   if (numtraces <= 0) {
      printf("\nERROR:  numtraces argument <= 0\n\n");
      exit(1);
   }

   int numsamples = atoi(argv[4]);
   if (numsamples <= 0) {
      printf("\nERROR:  number_samples_per_trace argument <= 0\n\n");
      exit(1);
   }

   double sampint = atof(argv[5]);
   if (sampint <= 0.0) {
      printf("\nERROR:  sample_interval argument <= 0.0\n\n");
      exit(1);
   }

   char dummy[3601];  // dummy buffer for skipping unused header information

   if ((in = fopen(argv[1],"r")) == NULL) {
      printf("\nERROR:  cannot open input xyz file %s\n\n",argv[1]);
      exit(1);
   }

   if ((out = fopen(argv[2],"w")) == NULL) {
      printf("\nERROR:  cannot open output SEGY file %s\n\n",argv[2]);
      exit(1);
   }

   // READ IN LINE UNTIL END OF FILE
   char string[200];

   int trace;
   int ptrace = -1;
   int ntrace = 0;

   double sample;
   double psample = -1.0;
   double firstsample = -1.0;
   double lastsample = -1.0;
   int nsamp = 0;

   float value;

   int flag = 1;

   // WRITE 3600 BYTE HEADER
   int i;
   short s;
   fwrite(dummy, 3200, 1, out);
   i = 1;
   fwrite(&i, 4, 1, out);
   fwrite(&i, 4, 1, out);
   fwrite(&i, 4, 1, out);
   s = (short)numtraces;
   fwrite(&s, 2, 1, out);
   s = 0;
   fwrite(&s, 2, 1, out);
   s = (short)(sampint * 1000.0);
   fwrite(&s, 2, 1, out);
   fwrite(&s, 2, 1, out);
   s = (short)numsamples;
   fwrite(&s, 2, 1, out);
   fwrite(&s, 2, 1, out);
   s = 5;
   fwrite(&s, 2, 1, out);
   fwrite(dummy, 374, 1, out);

   // READ XYZ FILE
   while(fgets(string, 200, in) != NULL) {

      if ( sscanf(string, "%d %lf %f\n", &trace, &sample, &value) != 3) {
         continue;
      }

      if (flag == 1) {
         firstsample = sample;
         flag = 2;
      }

      if (ptrace < 0) {
         ntrace++;
         ptrace = trace;
         writeheader(trace, numsamples, sampint, out);
      }

      if (trace != ptrace) {
       
         if (nsamp != numsamples) {
            printf("\nERROR:  %d samples in trace %d\n\n",nsamp, ptrace);
            exit(1);
         }
         nsamp = 0;

         ntrace++;

         writeheader(trace, numsamples, sampint, out);

         ptrace = trace;

         if (flag == 2) {
            lastsample = psample;
            flag = 0;
         }

         if (psample != lastsample) {
            printf("\nERROR:  last sample time was %f, now is %f\n\n",lastsample, psample);
            exit(1);
         }

         if (sample != firstsample) {
            printf("\nERROR:  first sample time was %f, now is %f\n\n",firstsample, sample);
            exit(1);
         }
      }

      else if (psample >= 0.0) {
         if ( fabs( sample - psample - (sampint/1000.0) ) > .001 ) {
            printf("\nERROR:  difference from sample %f to %f not equal to sample interval of %f ms\n\n",
               psample, sample, sampint/1000.0);
            exit(1);
         }
      }

      psample = sample;

      //write value
      fwrite(&value, 4, 1, out);

      nsamp++;

   }

   if (ntrace != numtraces) {
      printf("\nERROR:  only %d traces in input xyz file\n\n",ntrace);
      exit(1);
   }

   fclose(in);
   fclose(out);

}

void writeheader(int trace, int numsamples, double sampint, FILE* out) {
   char dummy[241];
   int i;
   short s;
   i = trace;
   fwrite(&i, 4, 1, out);
   fwrite(&i, 4, 1, out);
   i = 1001;
   fwrite(&i, 4, 1, out);
   i = trace;
   fwrite(&i, 4, 1, out);
   fwrite(dummy, 12, 1, out);
   s = 1;
   fwrite(&s, 2, 1, out);
   fwrite(dummy, 84, 1, out);
   s = (short)numsamples;
   fwrite(&s, 2, 1, out);
   s = (short)(sampint * 1000.0);
   fwrite(&s, 2, 1, out);
   fwrite(dummy, 122, 1, out);
}
